// bxscript.js
// Copyright (C) Manny Neira, 2017
// All rights reserved



// C O N S T A N T S

const BABLEX_VERSION = 'Bablex 1.0'
const SPECIAL_CHARS = '`\\*/_^|:<>[]{}&'



// TRANSLATOR PAGE CONTENT

const translator_page = `### Bablex Translator

This page will translate a text file marked up using //Bablex// into //HTML5// and then display it as a web page.

The controls are explained below.

+-
*PLEASE NOTE:*
//These controls can take time to process documents.//
//Don't worry if they don't respond immediately.//
-+

[Bablex Logo: bxlogo.gif] **The//Bablex// Logo:** Click the logo in the top right corner of the screen to hide (or reveal) the controls.

**File browser**: Use this to select the file to translate, which will usually have a \`.txt\` or \`.bx\` suffix.

**//Bablex// button**: Click this button to redisplay the original//Bablex// text.

**Contents button**: Click for a//Bablex// marked up outline of links to the headings in the document, which you may optionally cut and paste back into the top of your text as a contents listing.

**HTML button**: Click this to display the translated HTML. Hide the controls, select all this text, copy it, paste it into a text editor, and save with an \`.html\` suffix to create a file suitable for the web.

(Note that the HTML will contain the link: \`<link rel="stylesheet" href="bxstyle.css" type="text/css">\`, so either keep a copy of \`bxstyle.css\` with your \`.html\` file, or change this line to point to a different stylesheet. By all means copy \`bxstyle.css\` to create your own stylesheet, but it is best /not/ to change the original.)

**Page button**: Click this to return to the web page display.

=== More about Bablex ===

If you are new to//Bablex//, try opening \`Bablex Guide.txt\` for a detailed description.

And thank you for your interest in//Bablex//, and your use of **plain text** - the world's only /truly/ open format :)

+-
*//Bablex// - the markup language for people who love text!*
//Manny Neira// | me@mannyneira.com | [http://mannyneira.com]
-+`



// PARAGRAPH TYPE REGEXES

const PROTECTED_PARA_RX
  = /^ {1,3}[^ ](.|\n)*$/
const FORMATTED_PARA_RX
  = /^ {4}(.|\n)*$/
const PREFIXED_HEADING_RX
  = /^(#{3,}|={3,}|~{3,}|-{3,}|\.{3,}) +[^ ].*$/
const UNDERLINED_HEADING_RX
  = /^.+\n(#{3,}|={3,}|~{3,}|-{3,}|\.{3,})$/
const NUMBERED_LIST_RX
  = /^[0-9]+\. +[^ ].*(\n *([0-9]+\.){1,5} +[^ ].*)+$/
const BULLETED_LIST_RX
  = /^\* .*(\n *[\*\+\-] .*)+$/
const NUMBERED_ITEM_RX
  = /^[0-9]+\. [^ ].*$/
const BULLETED_ITEM_RX
  = /^\* [^ ].*$/
const TABLE_RX
  = /^\|.*\|(\n(\||:).*\|)+$/
const INSERT_OPEN_RX
  = /^(\+|")-+(\n|$)/
const INSERT_CLOSE_RX
  = /(^|\n)-+(\+|")$/
const RULE_RX
  = /^\*{3,}( [^ ].*)?$/
const IMAGE_RX
  = /^\[.*[^ ]+.*: +[^ ]+.*\].*$/
const TERM_RX
  = /^[^ ].* :: [^ ].*$/
const METADATA_RX
  = /^([^|]+ \| [^|]+(\n|$))+/



// OTHER REGEXES

const FILE_RX
  = /^.+\.(bx|txt)$/
const BOOKMARKED_LINK_RX
  = /^<.+> *\[.+\]$/
const AUTO_BOOKMARK_RX
  = /^(top|bottom|(heading-[0-9]+))$/
const FORMATTED_OPEN_RX
  = /^`-+(\n(.|\n)*)?$/
const FORMATTED_CLOSE_RX
  = /^((.|\n)*\n)?-+`$/



// G L O B A L   V A R I A B L E S

var source_file = 'bx2web.bx'
var input_text = ''
var input_paras = []
var output_type = 'page'
var output_text = ''
var error_message = ''



// DOCUMENT ANALYSIS VARIABLES

var main_heading = ''
var heading_count = 0
var heading_a_level = 0
var heading_b_level = 0
var heading_c_level = 0
var heading_d_level = 0
var heading_e_level = 0
var bookmark_names = []
var bookmark_targets = []
var term_patterns = []
var term_ids = []
var metadata_names = []
var metadata_values = []



// TRANSLATION WORKING VARIABLES

var para_ix = 0
var heading_ix = 0
var heading_stack = []
var insert_ix = 0
var insert_ix_stack = []
var insert_prefix_stack = []
var numbered_list_flag = false
var bulleted_list_flag = false



// P A G E   F U N C T I O N S



// INITIALISE ON LOADING THE TRANSLATOR PAGE.

window.onload = function ()
  {

// Check Javascript can read files.

  if (!(window.File && window.FileReader
    && window.FileList && window.Blob))
    {
    alert('Browser cannot read files by script.')
    return
    }

// Switch on controls and display translator page.

  set_controls(true)
  input_text = translator_page
  extract_paras()

// Make `get_chosen_file()` run when a file chosen.

  document.getElementById('file-input')
    .addEventListener('change',
      function (event) { get_chosen_file() }, false)
  }



// SET CONTROLS BY SETTING/RUNNING HIDDEN CLASS.

function set_controls(visible)
  {
  var controls_class = ''
  if (!visible)
    controls_class = 'hidden'
  document.getElementById('controls')
    .className = controls_class
  set_buttons()
  }

function toggle_controls()
  {
  set_controls(document.getElementById('controls')
    .className.indexOf('hidden') >= 0)
  }



// SET BUTTON CLASSES TO TRUE OR FALSE.

function set_buttons()
  {
  document.getElementById('bablex').className
    = (output_type == 'bablex')
  document.getElementById('contents').className
    = (output_type == 'contents')
  document.getElementById('html').className
    = (output_type == 'html')
  document.getElementById('page').className
    = (output_type == 'page')
  }

function toggle_button(element)
  {
  output_type = element.innerHTML
  set_buttons()
  display_output()
  }



// L O A D I N G   F U N C T I O N S


// GET THE CHOSEN FILE AND LOAD ITS CONTENT.

function get_chosen_file()
  {
  var file = document
    .getElementById('file-input').files[0]
  source_file = file.name
  if (!(FILE_RX.test(source_file)))
    {
    alert('Sorry, cannot translate "' + source_file
      + '", only ".bx" or ".txt" files.')
    return
    }
  var file_reader = new FileReader()
  file_reader.onload  = load_file_content
  file_reader.readAsText(file)
  }

function load_file_content(event)
  {
  set_controls(false)
  input_text = event.target.result
  extract_paras()
  }



// MERGE LINES INTO PARAGRAPHS.

function extract_paras()
  {

// Remove CRs and trailing whitespace/blank lines.

  var text = input_text
  text = text.replace(/\r/g, '')
  text = text.replace(/\t/g, '   ')
  text = text.replace(/[ \t]+\n/g, '\n')
  text = text.replace(/^\n*/, '')
  text = text.replace(/\n*$/, '')

// Treat lines as paragraaphs, then merge as req'd.

  input_paras = text.split('\n')

  para_ix = 0
  while (para_ix < input_paras.length - 1)
    {
    var next_ix = para_ix + 1
    var this_para = input_paras[para_ix]
    var next_para = input_paras[para_ix + 1]

// Merge 4 spaced lines and blanks between them.

    if (FORMATTED_PARA_RX.test(this_para))
      {
      if (FORMATTED_PARA_RX.test(next_para)
        || next_para.length == 0)
        {
        input_paras[para_ix] += '\n' + next_para
        input_paras.splice(para_ix + 1, 1)
        }
      else
        {
        input_paras[para_ix]
          = this_para.replace(/( |\n)*$/, '')
        para_ix += 1
        }
      }

// Merge all lines between `- and -`.

    else if (FORMATTED_OPEN_RX.test(this_para)
      && !FORMATTED_CLOSE_RX.test(this_para))
      {
      input_paras[para_ix] += '\n' + next_para
      input_paras.splice(para_ix + 1, 1)
      }

// If this line is blank, delete it.

    else if (this_para.length == 0)
      input_paras.splice(para_ix, 1)

// Treat insert boundaries (+- -+ "- -") as paras.

    else if (INSERT_OPEN_RX.test(this_para)
      || INSERT_CLOSE_RX.test(this_para)
      || INSERT_OPEN_RX.test(next_para)
      || INSERT_CLOSE_RX.test(next_para))
      para_ix += 1

// Join consecutive lines around a line break.

    else if (next_para.length > 0)
      {
      input_paras[para_ix] += '\n' + next_para
      input_paras.splice(para_ix + 1, 1)
      }

// Start new paragraphs after blank lines.

    else
      para_ix += 1
    }

// Display the loaded document in page view mode.

  output_type = 'page'
  display_output()
  }



// D I S P L A Y   F U N C T I O N S



// DISPLAY THE OUTPUT.

function display_output()
  {
  output_text = ''
  error_message = ''

// Call appropriate function to generate the output.

  if (output_type == 'bablex')
    output_bablex()
  else if (output_type == 'contents')
    output_contents()
  else if (output_type == 'html')
    output_html()
  else
    output_page()

// If an error was encountered, output that instead.

  if (error_message)
    output_error()

// Output by loading a page element with HTML.

  document.getElementById('output')
    .innerHTML = output_text
  }



// ADD TEXT TO THE OUTPUT BUFFER.

function add_output(text)
  {
  text = decoded(text)
  if (output_type == 'html')
    text = '<pre class="html"><code class="html">'
      + detagged(text) + '</code></pre><br class="html">'
  output_text += text + '\n';
  }



// OUTPUT THE LAST ERROR.

function output_error()
  {
  output_type = 'page'
  output_text = '<h1>Sorry...</h1>'
    + '<p>A problem arose during translation:</p>'
    + '<p><b><mark>' + error_message + '</mark></b></p>'

// Add context to the error message.

  var context_ix = para_ix - 10
  if (context_ix < 0)
    context_ix = 0
  var context_ends = context_ix + 20
  if (context_ends > input_paras.length)
    context_ends = input_paras.length
  output_text += '<h2>Context</h2>\n<pre>'
  while (context_ix < context_ends)
    {
    if (context_ix == para_ix)
      output_text += '<b><mark>'
    output_text += detagged(input_paras[context_ix])
    if (context_ix == para_ix)
      output_text += '</mark></b>'
    output_text += '\n<br>'
    context_ix += 1
    }
  output_text+= '</pre>'
  }



// OUTPUT THE ORIGINAL (IF NOW CLEANED UP) BABLEX.

function output_bablex()
  {
  add_output('<pre class="bablex"><code class="bablex">'
    + detagged(input_text) + '</code></pre>')
  }



// OUTPUT CONTENTS - A HEADING LINKS OUTLINE.

function output_contents()
  {
  add_output('<pre class="contents">'
    + '<code class="contents">*** CONTENTS ***\n')
  var counters = [0, 0, 0, 0, 0]
  var stack = ''
  var insert_depth = 0
  var count = input_paras.length
  var ix = 0
  while (ix < count)
    {
    var para = input_paras[ix]

// Track nested inserts depth.

    if (INSERT_OPEN_RX.test(para))
      insert_depth += 1
    else if (INSERT_CLOSE_RX.test(para))
      insert_depth -= 1

// Add headings (outside inserts) to contents list.

    else if (insert_depth == 0
      && (PREFIXED_HEADING_RX.test(para)
        || UNDERLINED_HEADING_RX.test(para)))
      {

// Extract title text and type from first char.

      var heading_prefix = ''
      var break_ix = para.indexOf('\n')
      if (break_ix < 0)
        {
        heading_prefix = para[0]
        var all_markers_regex = new RegExp('(^\\'
          + heading_prefix + '*|\\'
          + heading_prefix + '*$)', 'g')
        para = para.replace(all_markers_regex, '')
        }
      else
        {
        heading_prefix = para[break_ix + 1]
        para = para.slice(0, break_ix)
        }
      para = detagged(para.trim())
      var new_type = '#=~-.'.indexOf(heading_prefix)

// Remove equal/sub hdgs from stack & add this.

      while (true)
        {
        var old_type = -1
        if (stack)
          old_type = '#=~-.'.indexOf(stack.slice(-1)[0])
        if (new_type > old_type)
          break
        stack = stack.slice(0, -1)
        }
      stack += heading_prefix

// Note new heading depth & increment counter.

      var level = stack.length - 1
      counters[level] += 1

// Add heading to contents list.

      var prefix = ' '.repeat(level * 5)
      var part = 0
      while (part <= level)
        prefix += counters[part++] + '.'
      while (part < 5)
        counters[part++] = 0
      add_output(prefix + ' [' + para + ']')
      }

// Loop back and close contents list at end.

    ix += 1
    }
  add_output('</code></pre>')
  }



// OUTPUT HTML: TRANSLATE PAGE, BUT SHOW COMPLETE HTML.

function output_html()
  {
  output_page()

// Prepare comment block.

  var comment = '<!-- '
    + BABLEX_VERSION + ' Translation'
    + '\nSource | ' + source_file
    + '\nU.T.C. | ' + datetime()
    + '\nhttp://mannyneira.com/bablex -->'

// Prepare header block.

  var header = '<!DOCTYPE html>'
    + '\n<html lang="'
      + metadata_value('language', 'en') + '">'
    + '\n<head>\n<title>' + metadata_value('title',
      main_heading) + '</title>'
    + '\n<meta charset="'
      + metadata_value('charset', 'UTF-8') + '">'

// Add author/desc/keywords to header if specified.

  var count = metadata_names.length
  var ix = 0
  while (ix < count)
    {
    var name = metadata_names[ix]
    var value = metadata_values[ix]
    if (/^(author|description|keywords)$/.test(name))
      header += '\n<meta name="' + name
        + '" content="' + value + '">'
    ix += 1;
    }

// Finish header.

  header += '\n<meta name="generator" content="'
    + BABLEX_VERSION + '">'
    + '\n<link rel="stylesheet"'
      + ' href="bxstyle.css" type="text/css">'
    + '\n</head>\n<body>\n<main>'

// Prepare footer block.

  var footer = '</main>\n</body>\n</html>'

// Put together the HTML translation.

  output_text = '<pre class="html"><code class="html">'
      + detagged(comment)
      + '</code></pre><br class="html">'
    + '<pre class="html"><code class="html">'
      + detagged(header)
      + '</code></pre><br class="html">'
    + output_text
    + '\n<pre class="html"><code class="html">'
      + detagged(footer) + '</code></pre>'
  }



// OUTPUT THE TRANSLATED PAGE.

function output_page()
  {

// Phase 1 of 2: collect the following data.

  main_heading = ''
  heading_count = 0
  heading_a_level = 0
  heading_b_level = 0
  heading_c_level = 0
  heading_d_level = 0
  heading_e_level = 0
  bookmark_names = []
  bookmark_targets = []
  term_patterns = []
  term_ids = []
  metadata_names = []
  metadata_values = []

// Cycle through the paragraphs.

  var para_count = input_paras.length
  var heading_markers_used = ''
  insert_ix = 0
  para_ix = -1
  while (++para_ix < para_count)
    {
    var para = input_paras[para_ix]

// Don't analyse protected & formatted paras.

    if (PROTECTED_PARA_RX.test(para)
      || FORMATTED_PARA_RX.test(para)
      || FORMATTED_OPEN_RX.test(para))
      continue

// If this is a term, record the details.

    if (TERM_RX.test(para))
      {
      var colon_ix = para.indexOf(' :: ')
      var pattern = para.slice(0, colon_ix).trim()
      var pattern_id = id(pattern)
      var modifiers = ''
      if (pattern == pattern.toLowerCase())
        modifiers = 'i'
      pattern = term_pattern(pattern)
      term_patterns.push(new RegExp(pattern, modifiers))
      term_ids.push(pattern_id)
      }

// Track the insert depth.

    else if (INSERT_OPEN_RX.test(para))
      {
      insert_ix += 1
      insert_ix_stack.push(insert_ix)
      insert_prefix_stack.push(para[0])
      }

    else if (INSERT_CLOSE_RX.test(para))
      {
      if (insert_ix_stack.length == 0)
        {
        error_message = 'Insert closed when none open.'
        return
        }
      insert_ix_stack.splice(-1, 1)
      insert_prefix_stack.splice(-1, 1)
      }

// If this line is a heading, note & bookmark it.

    else if (PREFIXED_HEADING_RX.test(para)
      || UNDERLINED_HEADING_RX.test(para))
      {

// Extract the heading text.

      var heading_prefix = ''
      var break_ix = para.indexOf('\n')
      if (break_ix < 0)
        {
        heading_prefix = para[0]
        var all_markers_regex = new RegExp('(^\\'
          + heading_prefix + '*|\\'
          + heading_prefix + '*$)', 'g')
        para = para.replace(all_markers_regex, '')
        }
      else
        {
        heading_prefix = para[break_ix + 1]
        para = para.slice(0, break_ix)
        }
      para = para.trim()

// Count the headings in the main text.

      if (insert_ix_stack.length == 0)
        heading_count += 1

// Note the main heading.

      if (main_heading == ''
        && insert_ix_stack.length == 0)
        {
        if (heading_prefix != '#')
          {
          error_message
            = 'First heading not an A-Heading.'
          return
          }
        main_heading = para
        }

// Note heading levels used after main heading.

      else if (heading_markers_used
        .indexOf(heading_prefix) < 0)
        heading_markers_used += heading_prefix

// Note a bookmark for the heading.

      para = id(para)
      if (bookmark_names.indexOf(para) >= 0)
        {
        error_message = 'Heading bookmark "'
          + para + '" repeats.'
        return
        }
      else
        {
        bookmark_names.push(para)
        bookmark_targets.push('#' + para)
        }
      }

// Add <bookmark> [link] pair to the list.

    else if (BOOKMARKED_LINK_RX.test(para))
      {
      var bookmark = id(para.slice(1,
        para.indexOf('>')))
      if (bookmark_names.indexOf(bookmark) >= 0)
        {
        error_message = 'Bookmark "'
          + bookmark + '" repeats.'
        return
        }
      bookmark_names.push(bookmark)
      bookmark_targets.push(safe_url(
        para.slice(para.indexOf('[') + 1, -1)))
      }

// Look for embedded <bookmarks>.

    else
      {
      para = blankout(para, '\\', '\\')
      para = blankout(para, '`', '`')
      para = blankout(para, '[', ']')
      para = blankout(para, '{', '}')
      para = protected(para)
      var from_ix = 0
      var to_ix = 0
      while (true)
        {

// Translate any bookmarks in the paragraph.

        from_ix = para.indexOf('<', to_ix)
        to_ix = para.indexOf('>', from_ix + 1)
        if (from_ix < 0 || to_ix < 0)
          break
        if (para[from_ix + 1] == ' '
          || para[to_ix - 1] == ' ')
          continue
        var bookmark = id(para.slice(from_ix + 1,
          to_ix))
        if (bookmark_names.indexOf(bookmark) >= 0)
          {
          error_message = 'Bookmark "'
            + bookmark + '" repeats.'
          return
          }
        bookmark_names.push(bookmark)
        bookmark_targets.push('#'
          + id(para.slice(from_ix + 1, to_ix)))
        }
      }
    }

// Warn if not all inserts are closed.

  if (insert_ix_stack.length != 0)
    error_message = 'Not all the inserts were closed.'

// Choose heading levels for different heading types.

  var next_level = 2
  if (heading_markers_used.indexOf('#') >= 0)
    heading_a_level = next_level++
  if (heading_markers_used.indexOf('=') >= 0)
    heading_b_level = next_level++
  if (heading_markers_used.indexOf('~') >= 0)
    heading_c_level = next_level++
  if (heading_markers_used.indexOf('-') >= 0)
    heading_d_level = next_level++
  if (heading_markers_used.indexOf('.') >= 0)
    heading_e_level = next_level

// Phase 2 of 2: cycle through translating paras.

  heading_ix = 0
  heading_stack = []
  insert_ix = 0
  insert_ix_stack = []
  insert_prefix_stack = []
  numbered_list_flag = false
  bulleted_list_flag = false

// Open <header> for <main>, and mark as top.

  add_output(tag('!-- main header begins --')
    + '\n' + tag('header id="top"'))

// Loop round translating the paras.

  insert_ix = 0
  para_ix = 0
  while (para_ix < para_count)
    {
    var para = input_paras[para_ix]

    if (PROTECTED_PARA_RX.test(para))
      translate_protected_para(para)

    else if (FORMATTED_PARA_RX.test(para))
      translate_formatted_para(para)

    else if (FORMATTED_OPEN_RX.test(para))
      translate_formatted_block(para)

    else if (PREFIXED_HEADING_RX.test(para)
      || UNDERLINED_HEADING_RX.test(para))
      translate_heading(para)

    else if (RULE_RX.test(para))
      translate_rule(para)

    else if (NUMBERED_LIST_RX.test(para)
      || BULLETED_LIST_RX.test(para))
      translate_list(para)

    else if (NUMBERED_ITEM_RX.test(para)
      || BULLETED_ITEM_RX.test(para))
      translate_item(para)

    else if (TABLE_RX.test(para))
      translate_table(para)

    else if (IMAGE_RX.test(para))
      translate_image(para)

    else if (TERM_RX.test(para))
      translate_term(para)

    else if (METADATA_RX.test(para))
      translate_metadata(para)

    else if (INSERT_OPEN_RX.test(para))
      translate_insert_open(para)

    else if (INSERT_CLOSE_RX.test(para))
      translate_insert_close(para)

    else if (para)
      translate_plain_para(para)

// If there's an error, give up.

    if (error_message)
      return

// Loop back for the next para.

    para_ix += 1
    }

// Close any sections still open.

  var ix = heading_stack.length
  while (--ix >= 0)
    add_output(tag('!-- #'
      + heading_stack.slice(-1)[0].substring(1)
      + ' ends --') + '\n' + tag('/section'))

// Trap unclosed inserts or missing main heading.

  if (insert_ix_stack.length > 0)
    error_message = 'Not all the inserts were closed.'
  else if (main_heading == '')
    error_message = 'Document has no main heading.'

// Add a navigation block at the end of the document.

  add_output(tag('nav id="bottom"') + '\n'
    + tag('p class="navigation-bar"') + '\n'
    + tag('a href="#heading-' + heading_count + '"')
    + '&lt; LAST' + tag('/a') + '\n'
    + tag('a href="#top"') + '^ TOP'
    + tag('/a') + '\n' + tag('/p') + '\n' + tag('/nav'))
  }



// T R A N S L A T I O N   F U N C T I O N S



// TRANSLATE A PROTECTED PARAGRAPH.

function translate_protected_para(para)
  {

// Drop the first protective space & kill rest.

  add_output(tag('p')
    + dead(para.replace(/\n/g, tag('br')).slice(1))
    + tag('/p'))
  }



// TRANSLATE FORMATTED PARAGRAPH (4 SPACE PREFIX).

function translate_formatted_para(para)
  {
  add_output(tag('pre') + tag('code')
    + dead(para.replace(/\n*$/g, '')
      .replace(/(^|(\n)) {4}/g, '$2'))
    + tag('/code') + tag('/pre'))
  }



// TRANSLATE FORMATTED BLOCK (`--- TO ---`).

function translate_formatted_block(para)
  {
  if (!FORMATTED_CLOSE_RX.test(para))
    error_message
      = 'Unclosed formatted block: "-`" missing.'
  from = para.indexOf('\n')
  to = para.lastIndexOf('\n')
  if (from >= 0 && to >= from)
    add_output(tag('pre') + tag('code')
      + dead(para.slice(from + 1, to))
      + tag('/code') + tag('/pre'))
  }



// TRANSLATE A HEADING.

function translate_heading(para)
  {

// Collect para text, tag id, prefix & insert flag.

  var inserted = (insert_ix_stack.length != 0)
  var heading_prefix = ''
  var break_ix = para.indexOf('\n')
  if (break_ix < 0)
    {
    heading_prefix = para[0]
    var all_markers_regex = new RegExp('(^\\'
      + heading_prefix + '*|\\'
      + heading_prefix + '*$)', 'g')
    para = para.replace(all_markers_regex, '')
    }
  else
    {
    heading_prefix = para[break_ix + 1]
    para = para.slice(0, break_ix)
    }
  para = para.trim()
  var para_id = id(para)
  para = dead(para)

// Determine which HTML heading level to use.

  var new_marker_ix = '#=~-.'.indexOf(heading_prefix)
  var heading_level = [heading_a_level,
    heading_b_level, heading_c_level, heading_d_level,
    heading_e_level][new_marker_ix]

// Translate to dummy heading paragraph if an insert.

  if (inserted)
    {
    add_output(tag('p id="' + para_id + '" class="h'
      + heading_level + '-in-insert"') + para
      + tag('/p'))
    }

// Translate the main heading.

  else if (heading_ix == 0)
    {
    heading_ix = 1
    para = tag('h1 id="heading-1"')
        + tag('span id="' + para_id + '"')
        + para + tag('/span') + tag('/h1')
      + '\n' + tag('nav')
      + '\n' + tag('p class="navigation-bar"')
      + '\n' + tag('a href="#top"')
        + '^ TOP' + tag('/a')
      + '\n' + tag('a href="#bottom"')
        + 'v BOTTOM' + tag('/a')
    if (heading_count > 1)
      para += '\n' + tag('a href="#heading-2"')
        + 'NEXT &gt;' + tag('/a')
    para += '\n' + tag('/p') + '\n' + tag('/nav')
    add_output(para)
    add_output(tag('!-- main header ends --')
      + '\n' + tag('/header'))
    }

// Translate a subheading.

  else
    {
    heading_ix += 1

// Remove earlier menus and close sections.

      while (true)
      {
      var old_marker_ix = -1
      var old_marker_id = ''
      if (heading_stack.length > 0)
        {
        old_marker_ix = '#=~-.'
          .indexOf(heading_stack.slice(-1)[0][0])
        old_marker_id = heading_stack
          .slice(-1)[0].substring(1)
        }
      if (new_marker_ix > old_marker_ix)
        break
      add_output(tag('!-- #' + old_marker_id
        + ' ends --') + '\n' + tag('/section') + '\n')
      heading_stack.splice(-1, 1)
      }

// Add heading to stack and open section.

    heading_stack.push(heading_prefix + para_id)
    add_output(tag('!-- #' + para_id + ' begins --')
      + '\n' + tag('section class="level-'
      + heading_level + '"'))

// Add the heading.

    para = tag('header id="' + para_id + '"')
      + '\n' + tag('h' + heading_level
        + ' id="heading-' + heading_ix + '"')
        + para + tag('/h' + heading_level)
      + '\n' + tag('nav')
      + '\n' + tag('p class="navigation-bar"')
      + '\n' + tag('a href="#' + 'heading-'
        + (heading_ix - 1) + '"')
        + '&lt; LAST' + tag('/a')
      + '\n' + tag('a href="#top"')
        + '^ TOP' + tag('/a')
      + '\n' + tag('a href="#bottom"')
        + 'v BOTTOM' + tag('/a')
    if (heading_ix < heading_count)
      para += '\n' + tag('a href="#heading-'
        + (heading_ix + 1) + '"')
        + 'NEXT &gt;' + tag('/a')
    para += '\n' + tag('/p') + '\n' + tag('/nav')
      + '\n' + tag('/header')
    add_output(para)
    }
  }



// TRANSLATE A RULE.

function translate_rule(para)
  {
  para = para.replace(/\*/g, '').trim()
  if (para)
    add_output(tag('hr class="rule-with-text"')
      + '\n' + tag('p class="rule-with-text"')
      + dead(para) + tag('/p'))
  else
    add_output(tag('hr'))
  }



// TRANSLATE A LIST.

function translate_list(para)
  {

// Determine what type of list this is.

  var numbered = (para[0] != '*')
  var list_open_tag = tag('ol')
  var list_close_tag = tag('/ol')
  if (!numbered)
    {
    list_open_tag = tag('ul')
    list_close_tag = tag('/ul')
    }

// Loop round the items in the list.

  var lines = para.split('\n')
  para = list_open_tag
  var level = 1
  var count = lines.length
  var ix = 0
  while (ix < count)
    {
    var line = lines[ix++].trim()

// Get the prefix and level.

    var space_ix = line.indexOf(' ')
    var prefix = line.slice(0, space_ix)
    line = live(line.slice(space_ix + 1))
    var new_level = '*+-'.indexOf(prefix) + 1
    if (new_level <= 0)
      new_level = prefix.split('.').length - 1

// Determine if the item has a sublist.

    var has_sublist = false
    if (ix < count)
      {
      var next_line = lines[ix].trim()
      var next_prefix = next_line.slice(0,
        next_line.indexOf(' '))
      var next_level = '*+-'.indexOf(next_prefix) + 1
      if (next_level <= 0)
        next_level = next_prefix.split('.').length - 1
      if (new_level < next_level)
        has_sublist = true
      }

// Report skipped levels.

    if (new_level > level + 1)
      {
      error_message = 'Level skipped in list.'
      return
      }

// If the level has changed, open or close lists.

    while (new_level > level)
      {
      para += '\n' + list_open_tag
      level += 1
      }
    while (new_level < level)
      {
      para += '\n' + list_close_tag + '\n' + tag('/li')
      level -= 1
      }

// Add the item (don't close if it has a sublist).

    para += '\n' + tag('li') + line
    if(!has_sublist)
      para += tag('/li')
    }

// Close any lists/items still open.

  while (--new_level > 0)
    para += '\n' + list_close_tag + '\n' + tag('/li')
  para += '\n' + list_close_tag
  add_output(para)
  }



// TRANSLATE AN ISOLATED LIST ITEM.

function translate_item(para)
  {
  var numbered = (para[0] != '*')
  para = tag('li class="isolated-item"')
    + live(para.slice(para.indexOf(' ') + 1))
    + tag('/li')

// If this is the first item, open & flag the list.

  if (numbered && !numbered_list_flag)
    {
    para = tag('ol') + para
    numbered_list_flag = true
    }
  else if (!numbered && !bulleted_list_flag)
    {
    para = tag('ul') + para
    bulleted_list_flag = true
    }

// If this is the last item, close the list.

  var last_item = true
  if (para_ix < input_paras.length - 1)
    {
    var next_para = input_paras[para_ix + 1].trim()
    if (numbered
      && /^[0-9]+\. +[^ ].*$/.test(next_para))
      last_item = false
    else if (!numbered && next_para[0] == '*')
      last_item = false
    }
  if (last_item)
    {
    if (numbered)
      {
      para += tag('/ol')
      numbered_list_flag = false
      }
    else
      {
      para += tag('/ul')
      bulleted_list_flag = false
      }
    }

// Add the translated item.

  add_output(para)
  }



// TRANSLATE A TABLE.

function translate_table(para)
  {
  var rows = para.split('\n')
  var row_flags = []
  var row_count = rows.length
  var row_ix = 0
  var cols = []
  var col_count = 0
  var col_ix = 0

// Find the bars separating columns.

  while (row_ix < rows.length)
    {

// Create a protected copy of the row.

    row = rows[row_ix++]
    row = blankout(row, '```', '```')
    row = blankout(row, '``', '``')
    row = blankout(row, '`', '`')
    row = blankout(row, '\\\\\\', '\\\\\\')
    row = blankout(row, '\\\\', '\\\\')
    row = blankout(row, '\\', '\\')

// Note column bars, and which rows contain data.

    var bar_count = 0
    var col = 0
    while ((col = row.indexOf('|', col)) >= 0)
      {
      bar_count += 1
      if (cols.indexOf(col) < 0)
        cols.push(col)
      col += 1
      }
    row_flags.push(bar_count > 1
      && !/^(\||:| )*$/.test(row))
    }

// Sort and count the columns.

  cols.sort(difference)
  col_count = cols.length - 1

// Loo round the rows.

  var first_row_length = rows[0].length
  var totals = false
  row_ix = 0
  while (row_ix < rows.length)
    {
    row = rows[row_ix]
    var new_row = ''

// Table rows must be the same length.

    if (row.length != first_row_length)
      {
      error_message = 'Table rows are not the same length.'
      return
      }

// Handle lines of dashes under column headings.

    if (/^(\|\-+)+\|$/.test(row))
      {
      if (row_ix > 0)
        {
        var row_above = rows[row_ix - 1]
        while (row_above.indexOf(encoded('<td')) >= 0)
          row_above = row_above
            .replace(encoded('<td'), encoded('<th'))
        while (row_above.indexOf(encoded('</td')) >= 0)
          row_above = row_above
            .replace(encoded('</td'), encoded('</th'))
        rows[row_ix - 1] = row_above
        }
      }

// If dotted line, set totals flag for next row.

    else if (/^(\|\.+)+\|$/.test(row))
      totals = true

// If the row contains data, translate it.

    else if (row_flags[row_ix])
      {
      new_row += '\n' + tag('tr')

// Loop columns & process cells beginning in each.

      col_ix = 0
      while (col_ix < col_count)
        {
        col = cols[col_ix]
        if (row[col] == '|')
          {

// Determine the horizontal span of the cell.

          var hspan = 1
          while (col_ix + hspan < col_count
            && '|:'.indexOf(row[cols[col_ix + hspan]])
              < 0)
            hspan += 1

// Determine the vertical span of the cell.

          var vspan = 1
          var flagged_vspan = 1
          while (row_ix + vspan < rows.length
            && rows[row_ix + vspan][col] == ':')
            {
            if (row_flags[row_ix + vspan])
              flagged_vspan += 1
            vspan += 1
            }

// Extract the cell fro m the row.

          var end_col = cols[col_ix + hspan]
          var cell = row.slice(col + 1, end_col)
          var old_length = cell.length
          cell = cell.replace(/^ */, '')

// Determine the cell's alignment.

          var left_space = old_length - cell.length
          cell = cell.replace(/ *$/, '')
          var right_space = old_length
            - cell.length - left_space
          var alignment = 'left-cell'
          if (left_space > 1 && right_space > 1)
            alignment = 'centre-cell'
          else if (left_space > right_space)
            alignment = 'right-cell'

// If cell extends downwards, extract data.

          var ix = 0
          while (++ix < vspan)
            cell += '\n' + rows[row_ix + ix]
              .slice(col + 1, end_col).trim()
          cell = cell.replace(/\n\n/g, '\n')
            .replace(/\n/g, tag('br'))

// Translate the cell.

          classes = alignment
          if (totals)
            classes += ' total-cell'
          var td = 'td class="' + classes + '"'
          if (hspan > 1)
            td += ' colspan=' + hspan
          if (flagged_vspan > 1)
            td += ' rowspan=' + flagged_vspan
          new_row += '\n' + tag(td)
            + live(protected(cell)) + tag('/td')

// Loop back for next cell/next row.

          }
        col_ix += 1
        }
      new_row += '\n' + tag('/tr')
      totals = false
      }
    rows[row_ix] = new_row
    row_ix += 1
    }

// Wrap up and add the table paragraph.

  para = tag('table')
  row_ix = -1
  while (++row_ix < rows.length)
    if (rows[row_ix].length > 0)
      para += rows[row_ix]
  para += '\n' + tag('/table')
  add_output(para)
  }



// TRANSLATE AN IMAGE.

function translate_image(para)
  {
  var colon_ix = para.indexOf(': ')
  var closing_ix = para.indexOf(']')
  var caption = para.slice(1, colon_ix).trim()
  var alt = 'IMAGE_RX: ' + dead(caption)
  caption = live(caption)
  var link_para = para.slice(colon_ix + 1,
    closing_ix).trim()
  var link_url = safe_url(link_para)

// If link refers to a bookmark, use its destination.

  if (bookmark_names.indexOf(id(link_para)) >= 0)
    link_url = bookmark_targets[bookmark_names
      .indexOf(id(link_para))]

// Add the tags for the image.

  var para_text = ''
  if (closing_ix < para.length - 1)
    para_text = live(para.slice(closing_ix + 1).trim())
  var is_figure = (para_text == '')
  if (is_figure)
    para = tag('figure')
  else
    para = tag('p class="paragraph-with-image"')

  para += tag('span class="image-frame"')
    + tag('a href="' + link_url + '"')
    + tag('img src="' + link_url + '" alt="'
    + alt + '"') + tag('/a')

  if (!is_figure)
    para += tag('br') + caption

  para += tag('/span')

  if (is_figure)
    para += tag('figcaption') + caption + tag('br')
      + tag('a href="' + link_url + '"')
      + dead(link_para) + tag('/a')
      + tag('/figcaption') + tag('/figure')
  else
    para += tag('/p')
      + '\n' + tag('p') + para_text + tag('/p')

// Add the translation.

  add_output(para)
  }



// TRANSLATE A TERM DEFINITION.

function translate_term(para)
  {
  var term = para.slice(0, para.indexOf(' :: ')).trim()
  var definition = para.slice(para.indexOf(' :: ') + 4)
  para = tag('dl id="' + id(term)
    + '-term" class="term"')
    + '\n' + tag('dt') + dead(term) + tag('/dt')
    + '\n' + tag('dd') + live(definition) + tag('/dd')
    + '\n' + tag('/dl')
  add_output(para)
  }



// TRANSLATE METADATA.

function translate_metadata(para)
  {
  var lines = para.split('\n')
  para = tag('dl class="metadata"') + '\n'

// Loop round the metadata lines in the paragraph.

  var count = lines.length
  var ix = 0
  while (ix < count)
    {
    var line = lines[ix]
    var name = line.slice(0, line.indexOf('|')).trim()
    line = line.slice(line.indexOf('|') + 1).trim()

// Reject attempt to set generator different value.

    if (name.toLowerCase() == 'generator'
      && line.toLowerCase()
        != BABLEX_VERSION.toLowerCase())
      {
        error_message = 'Generator is "'
          + BABLEX_VERSION + '", not "' + line + '".'
        return
      }

// Translate the metadata.

    para += tag('dt') + dead(name) + tag('/dt') + '\n'
      + tag('dd') + live(line) + tag('/dd') + '\n'

// Reject multiple definitions of same property.

    if (insert_ix_stack.length == 0)
      {
      if (metadata_names.indexOf(name.toLowerCase) >= 0)
        {
        error_message = 'Metadata "'
          + name + '" repeated.'
        return
        }

// Record the metadata property and value.

      metadata_names.push(name.toLowerCase())
      metadata_values.push(line)
      }
    ix += 1
    }
  para += tag('/dl')
  add_output(para)
  }



// TRANSLATE INSERT OPENING AND CLOSURE.

function translate_insert_open(para)
  {
  insert_ix += 1
  insert_ix_stack.push(insert_ix)
  insert_prefix_stack.push(para[0])
  if (para[0] == '+')
    para = tag('div class="insert"')
  else
    para = tag('blockquote')
  para = '<!-- insert ' + insert_ix + ' begins -->\n' + para
  add_output(para)
  }

function translate_insert_close(para)
  {

// Reject closure when no insert open.

  if (insert_ix_stack.length == 0)
    {
    error_message = 'Insert closed when none open.'
    return
    }

// Note insert ix & prefix before removing from stack.

  var this_insert_ix = insert_ix_stack.slice(-1)[0]
  var this_insert_prefix = insert_prefix_stack.slice(-1)[0]
  insert_ix_stack.splice(-1, 1)
  insert_prefix_stack.splice(-1, 1)

// Reject attempt to close insert with wrong syntax.

  if (this_insert_prefix != para.slice(-1)[0])
    {
    error_message = 'Mismatched insert open & close.'
    return
    }

// Close the insert.

  para = '<!-- insert ' + this_insert_ix + ' ends -->\n'
  if (this_insert_prefix == '+')
    para += tag('/div')
  else
    para += tag('/blockquote')
  add_output(para)
  }



// TRANSLATE AN ORDINARY PARAGRAPH.

function translate_plain_para(para)
  {
  add_output(tag('p') + live(para) + tag('/p'))
  }



// E M B E D D I N G   F U N C T I O N S



// TRANSLATE MARKUP IN LIVE TEXT, PROTECT IN DEAD.

function live(text)
  {
  text = protected(text)
  text = termed(text)
  text = html_embedded(text)
  text = bookmarked(text)
  text = linked(text)
  text = styled(text)
  text = linebroken(text)
  text = detagged(text)
  return text
  }

function dead(text)
  {
  text = detagged(text)
  text = encoded(text)
  return text
  }



// SWAP BABLEX/HTML SYNTAC TO/FROM CODES FROM 255 DOWN.

function encoded(text)
  {
  var ix = SPECIAL_CHARS.length
  while (--ix >= 0)
  	text = text.replace(new RegExp('\\'
      + SPECIAL_CHARS[ix], 'g'),
      String.fromCharCode(255 - ix))
  return text
  }

function decoded(text)
  {
  var ix = SPECIAL_CHARS.length
  while (--ix >= 0)
    text = text.replace(new RegExp('\\'
      + String.fromCharCode(255 - ix), 'g'),
      SPECIAL_CHARS[ix])
  return text
  }



// KILL \PROTECTED\ AND `MONOSPACED` SPANS.

function protected(text)
  {
  while (true)
    {

// Protect using 3x, then 2x, then single markers.

    var level = 3
    var bits = null
    var code = false
    while (level > 0)
      {

// Built \\\...\\\ or ```...``` regexes.

      var pattern = new RegExp('(^(.|\n)*?)('
        + '\\\\'.repeat(level)
        + '([^ ]|[^ ](.|\n)*?[^ ])'
        + '\\\\'.repeat(level) + ')((.|\n)*$)')
      var code_pattern = new RegExp('(^(.|\n)*?)('
        + '`'.repeat(level) + '([^ ]|[^ ](.|\n)*?[^ ])'
        + '`'.repeat(level) + ')((.|\n)*$)')

// Apply whichever matches first in the string.

      bits = text.match(pattern)
      var code_bits = text.match(code_pattern)
      if (code_bits != null)
        {
        if (bits == null)
          code = true
        else if (bits[1].length > code_bits[1].length)
          code = true
        }

// If code regex (``) was found, swap the results.

      if (code)
        bits = code_bits

// If either regex found, break out to apply it.

      if (bits != null)
        break

// Loop back and try shorter markers.

      level -= 1
      }

// If bits is empty, there's nothing let to protect.

    if (bits == null)
      break

// Protect the but between the markers (if any).

    var middle = bits[3]
    if (middle.length <= (level * 2))
      middle = ''
    else
      {
      middle = dead(middle.slice(level, -level))
      if (code)
        middle = tag('code') + middle + tag('/code')
      }
    text = bits[1] + middle + bits[6]
    }
  return text
  }



// RETURN TEXT WITH TERMS LINKED TO DEFINITIONS.

function termed(text)
  {
  var copy = text

// Prevent terms being processed within certain tokens.

  copy = blankout(copy, '\\', '\\')
  copy = blankout(copy, '`', '`')
  copy = blankout(copy, '<', '>')
  copy = blankout(copy, '[', ']')
  copy = blankout(copy, '{', '}')

// Loop until no terms remain to be linked.

  var term_count = term_patterns.length
  while (true)
    {
    var last_term = ''
    var last_term_ix = -1
    var last_find_ix = -1

// Find which term is furthest to the right.

    var term_ix = term_count
    while (--term_ix >= 0)
      {
      var bits = copy.match(term_patterns[term_ix])
      if (bits)
        {
        var find_ix = bits[1].length
        if (find_ix > last_find_ix)
          {
          last_term = bits[4]
          last_term_ix = term_ix
          last_find_ix = find_ix
          }
        }
      }

// Link up rightmost term if one was found.

    if (last_term_ix < 0)
      break
    text = text.slice(0, last_find_ix)
      + tag('a class="term" href="#'
      + term_ids[last_term_ix] + '-term"') + last_term
      + tag('/a')
      + text.slice(last_find_ix + last_term.length)
    copy = copy.slice(0, last_find_ix)
    }
  return text
  }

// Return text with {HTML} embedded.

function html_embedded(text)
  {
  var from_ix = 0
  var to_ix = 0
  new_text = ''
  while (true)
    {
    from_ix = text.indexOf('{', 0)
    to_ix = text.indexOf('}', from_ix + 1)
    if (from_ix < 0 || to_ix < 0)
      break
    if (text[from_ix + 1] == ' '
      || text[to_ix - 1] == ' ')
      new_text += text.slice(0, to_ix + 1)
    else
      {
      var code = encoded(text.slice(from_ix + 1, to_ix),
        false)
      new_text += text.slice(0, from_ix)
        + code.replace(/\n/g, ' ')
      }
    text = text.slice(to_ix + 1)
    }
  return new_text + text
  }



// RETURN TEXT WITH EMBEDDED <BOOKMARKS>.

function bookmarked(text)
  {

// Make map of text excluding no-change zones.

  var copy = text
  copy = blankout(copy, '\\', '\\')
  copy = blankout(copy, '`', '`')
  copy = blankout(copy, '[', ']')
  copy = blankout(copy, '{', '}')

// Replace bookmarks.

  var from_ix = 0
  var to_ix = 0
  new_text = ''

// Loop until all bookmarks replaced.

  while (true)
    {
    from_ix = copy.indexOf('<', 0)
    to_ix = copy.indexOf('>', from_ix + 1)

// None find - break out.

    if (from_ix < 0 || to_ix < 0)
      break

// Protective spaces inside <>, don't change.

    if (text[from_ix + 1] == ' '
      || text[to_ix - 1] == ' ')
      new_text += text.slice(0, to_ix + 1)

// Live bookmark found: process.

    else
      {
      var bookmark = text.slice(from_ix + 1, to_ix)
      var bookmark_tag = 'span id="'
        + id(bookmark) + '"'

// Reject clashes with automatic bookmarks.

      if (AUTO_BOOKMARK_RX.test(id(bookmark)))
        {
        error_message = 'Bookmark "' + bookmark
          + '" clashes with automatic bookmark.'
        return ''
        }

// A numeric bookmark is a footnote.

      else if (/^[0-9]+$/.test(bookmark))
        bookmark_tag += ' class="footnote"'

// A valid bookmark name.

      else if (/^([a-zA-Z][0-9a-zA-Z \-]*|[0-9]+)$/
        .test(bookmark))
        bookmark_tag += ' class="bookmark"'

// Reject invalid bookmark names.

      else
        {
        error_message = 'Bookmarks must begin with'
          + ' a letter and contain only letters,'
          + ' numbers, spaces and dashes, OR be'
          + ' entirely numeric, unlike "'
          + bookmark + '".'
        return ''
        }

// Wrap the bookmark in a link.

      new_text += text.slice(0, from_ix)
        + tag(bookmark_tag) + bookmark + tag('/span')
      }

// Throw away the text which has been checked.

    text = text.slice(to_ix + 1)
    copy = copy.slice(to_ix + 1)
    }

// Return the new text, and what is left of the old.
  return new_text + text
  }



// RETURN TEXT WITH EMBEDDED [LINKS].

function linked(text)
  {
  var from_ix = 0
  var to_ix = 0
  new_text = ''

// Loop until all links translated.

  while (true)
    {
    from_ix = text.indexOf('[', 0)
    to_ix = text.indexOf(']', from_ix + 1)

// No more links, all done.

    if (from_ix < 0 || to_ix < 0)
      break

// Protective spaces inside [], don't change text.

    if (text[from_ix + 1] == ' '
      || text[to_ix - 1] == ' ')
      new_text += text.slice(0, to_ix + 1)

// Live link found - translate it.

    else
      {
      var link = text.slice(from_ix + 1, to_ix).trim()
      var link_text = dead(link)
      var bookmark_ix = bookmark_names.indexOf(id(link))

// Identify the target of the link.

      if (bookmark_ix >= 0)
        link = bookmark_targets[bookmark_ix]
      else if (AUTO_BOOKMARK_RX.test(id(link)))
        link = '#' + id(link)


// Unless file/web links, reject if no bookmark.

      else if (!/\.[^ ]/.test(link))
        {
        error_message = 'Link "' + link
          + '" has no corresponding bookmark.'
        return ''
        }

// Handle file/web links.

      else
        {

// Convert local txt/bx files to html.

        if (link.indexOf('://') < 0)
          {
          if (/\.(txt|bx)$/.test(link))
            link = link.slice(0, link.lastIndexOf('.'))
              + '.html'
          if (/\.(txt|bx)$/.test(link_text))
            link_text = link_text.slice(0,
              link_text.lastIndexOf('.')) + '.html'
          }

// Translate foreign links to #fragments.

        var break_ix = link.indexOf('; ')
        if (break_ix >= 0)
          link = link.slice(break_ix + 2) + '#'
            + id(link.slice(0, break_ix))
        link = safe_url(link)
        }

// Translate the link in the new text.

      var link_tag = 'a href="' + link + '"'
      if (/^[0-9]+$/.test(link_text))
        link_tag += ' class="footnote"'
      new_text += text.slice(0, from_ix)
        + tag(link_tag) + link_text + tag('/a')
      }

// Drop already translated text.

    text = text.slice(to_ix + 1)
    }

// Return the new text.

  return new_text + text
  }



// RETURN TEXT WITH <tags> AND &chars; ESCAPED.

function detagged(para)
  {
  return para.replace(/&/g, '&amp;')
    .replace(/</g, '&lt;').replace(/>/g, '&gt;')
  }



// RETURN TEXT WITH */_^STYLES^_/* MARKED UP.

function styled(para)
  {
  para = single_styled(para, '**', tag('b')
    + tag('mark'), tag('/mark') + tag('/b'))
  para = single_styled(para, '//', tag('i')
    + tag('mark'), tag('/mark') + tag('/i'))
  para = single_styled(para, '__', tag('u')
    + tag('mark'), tag('/mark') + tag('/u'))
  para = single_styled(para, '*', tag('b'),
    tag('/b'))
  para = single_styled(para, '/', tag('i'),
    tag('/i'))
  para = single_styled(para, '_', tag('u'),
    tag('/u'))
  para = single_styled(para, '^', tag('mark'),
    tag('/mark'))
  return para
  }

function single_styled(para, symbol, opener, closer)
  {
  symbol = symbol.replace(/(.)/g, '\\$1')
  var span = new RegExp(symbol + '([^ ]|[^ ].*?[^ ])'
    + symbol, 'g')
  return para.replace(span, opener + '$1' + closer)
  }



// RETURN A PARAGRAPH BROKEN INTO LINES.

function linebroken(para)
  {
  return para.replace(/\n/g, '\n' + tag('br'))
  }



// MISCELLANEOUS FUNCTIONS



function tag(content)
  {
  return encoded('<' + content + '>', false)
  }

function safe_url(url)
  {
  return encodeURI(url).replace(/_/g, '%5F')
    .replace(/\*/g, '%2A')
  }

function metadata_value(name, default_value)
  {
  var ix = metadata_names.indexOf(name.toLowerCase())
  if (ix < 0)
    return default_value
  return metadata_values[ix]
  }

function term_pattern(pattern)
  {
  var matches = pattern.match(/\((.+)\)$/)

// This term has alternative endings in parentheses.

  if (matches)
    {
    pattern = pattern.slice(0, pattern.lastIndexOf('('))
    var endings = matches[1].split('/')
    var endings_count = endings.length
    var endings_ix = 1

// 1st ending null, so endings replace characters.

    if (endings[0] == '')
      {
      var endings_size = 0
      while (endings_ix < endings_count)
        {
        var size = endings[endings_ix++].length
        if (size > endings_size)
          endings_size = size
        }
      endings[0] = pattern.slice(-endings_size)
      endings_ix = 0
      while (++endings_ix < endings_count)
        endings[endings_ix] = (pattern
          + endings[endings_ix]).slice(-endings_size)
      pattern = pattern.slice(0, -endings_size)
      }

// Put alternative endings in (a|b|c) group.

    pattern += '('
    endings_ix = 0
    while (endings_ix < endings_count)
      {
      if (endings_ix > 0)
        pattern += '|'
      pattern += endings[endings_ix++]
      }
    pattern += ')'

// If only one ending, it's simply optional.

    if (endings_count == 1)
      pattern += '?'
    }

// No endings, but could be a/b/c options.

  else if (/.+(\/.+)+/.test(pattern))
    pattern = pattern.replace(/\//g, '|')

// Return a regex pattern which should find the term.

  return '(^(.|\\n)*(_|\\b))(' + pattern + ')(_|\\b)'
  }

function id(para)
  {
  para = para.toLowerCase().trim()
  para = para.replace(/[^a-z0-9]/g, '-')
  para = para.replace(/\-*(.*[^\-])\-*/, '$1')
  para = para.replace(/\-{2,}/g, '-')
  if (/^[0-9]+$/.test(para))
    {
    para = 'footnote-' + para
    if (insert_ix_stack.length > 0)
      para = 'insert-' + insert_ix_stack.slice(-1)[0]
        + '-' + para
    }
  else if (para == null)
    para = 'bad-id'
  return para
  }

function blankout(copy, from, to)
  {
  var symbol_length = (from + to).length
  from = from.replace(/(.)/g, '\\$1')
  to = to.replace(/(.)/g, '\\$1')
  var pattern = new RegExp('(^(.|\n)*)(' + from
    + '([^ ]|[^ ](.|\n)*?[^ ])' + to + ')')
  while (true)
    {
    var bits = copy.match(pattern)
    if (bits == null)
      break
    var prefix_length = bits[1].length
    var match_length = bits[4].length + symbol_length
    copy = copy.slice(0, prefix_length)
      + '-'.repeat(match_length)
      + copy.slice(prefix_length + match_length)
    }
  return copy
  }

function difference(a, b)
  {
  return a - b
  }

function datetime()
  {
  var now = new Date()
  return now.getUTCFullYear()
    + '-' + twodigit(now.getUTCMonth() + 1)
    + '-' + twodigit(now.getUTCDate())
    + ' ' + twodigit(now.getUTCHours())
    + ':' + twodigit(now.getUTCMinutes())
    + ':' + twodigit(now.getUTCSeconds())
  }

function twodigit(number)
  {
  if (number < 10)
    number = '0' + number
  return number
  }
